"""
Basic command handling in MoaT-RPC.
"""

from __future__ import annotations

from ._base import BaseCmd as BaseCmd
from ._base import LoadCmd as LoadCmd
from ._base import LockBaseCmd as LockBaseCmd
from ._base import RootCmd as _RootCmd

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from moat.lib.path import Path


class RootCmd(_RootCmd):
    """
    This is the system's root dispatcher.

    This class ducktypes :py:class:`~moat.lib.rpc.cmd._base.BaseCmd`.
    """

    def cfg_at(self, p: Path):
        "returns a CfgStore object at this subpath"
        from moat.lib.rpc.cmd.tree.dir import CfgStore  # noqa:PLC0415

        return CfgStore(self, p)



"""
Server-side dir support
"""

from __future__ import annotations

from itertools import chain

from moat.util import NotGiven
from moat.lib.path import Path

from ._dir import BaseSubCmd as BaseSubCmd
from ._dir import BaseSuperCmd as BaseSuperCmd
from ._dir import DirCmd as DirCmd

# Typing
from typing import TYPE_CHECKING  # isort:skip

if TYPE_CHECKING:
    from moat.lib.rpc import MsgSender


class _NotGiven:
    # This is distinct from the "real" NotGiven. This is intentional.
    pass


class SubStore:
    """
    A helper class to retrieve a possibly-encoded object.
    """

    cfg: dict = None

    def __init__(self, sd: MsgSender):
        self.__sd = sd

    async def __aenter__(self):
        return self

    async def __aexit__(self, *tb):
        pass

    async def get(self, *a, **kw):
        """
        Collect a parts-encoded reply.
        """
        rp = kw.pop("p", Path())

        async def _get(p):
            d = await self.__sd(*a, p=p, **kw)
            if isinstance(d, (list, tuple)) and len(d) == 2:
                d, s = d
                for k in s:
                    d[k] = await _get(p + (k,))
            return d

        return await _get(rp)


class CfgStore(SubStore):
    """
    Config file storage.

    The subpath points to the destination's "cfg.Cmd" app, or a
    similar class that exposes ``r``, ``w`` and possibly ``x`` commands.
    """

    def __init__(self, dispatch, path: Path):
        self.sd = dispatch.sub_at(path)
        super().__init__(self.sd.r)

    async def get(self, again=False):  # noqa: D102
        self.cfg = await super().get()
        if self.cfg and not again:
            return self.cfg
        cfg = await super().get()
        self.cfg = cfg
        return cfg

    async def set(self, cfg: dict, sync: bool = True, replace: bool = False, keep: bool = False):
        """
        Update the client's configuration data.

        Args:
            cfg: new config data
            keep: `NotGiven` values will be kept
            replace: delete existing config
        """

        async def _set(p, c, empty: bool = False):
            # current client cfg
            if not empty:
                try:
                    ocd = await self.sd.r(p=p)
                    if isinstance(ocd, (list, tuple)):
                        ocd, ocl = ocd
                    else:
                        ocl = ()
                except (KeyError, IndexError):
                    empty = True
            if empty:
                ocd = {}
                ocl = []
            for k, v in c.items() if isinstance(c, dict) else enumerate(c):
                if isinstance(k, str) and k.startswith("_"):
                    continue
                if isinstance(v, (dict, list)):
                    if isinstance(c, list) or isinstance(ocd, list):
                        if isinstance(ocd, dict) and not empty:
                            await _set(p, [])
                            ocd = []
                        if k is None:
                            k = len(ocd)  # noqa:PLW2901
                    if empty or not v:
                        await self.sd.c(p=p + (k,), d=type(v)(), keep=keep)
                    if v:
                        await _set(p + (k,), v, empty=empty)
                elif isinstance(ocd, dict):
                    if ocd.get(k, _NotGiven) != v:
                        await self.sd.c(p=p + (k,), d=v, keep=keep)
                elif isinstance(ocd, list):
                    if len(ocd) <= k or ocd[k] != v:
                        await self.sd.c(p=p + (k,), d=v, keep=keep)
                else:
                    raise ValueError(f"Orig is {ocd!r}")  # noqa:TRY004

            if not replace:
                return
            # drop those client cfg snippets that are not on the server
            for k in chain(ocd.keys(), ocl):
                if k not in c:
                    await self.sd.c(p=p + (k,), d=NotGiven)

        self.cfg = None
        await _set(Path(), cfg)

        if sync:
            await self.sd.x()  # runs



"""
Stream link-up support for MoaT commands
"""

from __future__ import annotations

from moat.util import merge
from moat.lib.codec.errors import SilentRemoteError
from moat.lib.micro import AC_use, BaseExceptionGroup, L, TaskGroup, idle, log  # noqa:A004
from moat.lib.rpc import BaseCmd, HandlerStream, NotReadyError

__all__ = ["BaseCmdMsg", "CmdMsg", "ExtCmdMsg", "MsgStream", "SingleCmdMsg"]

# Typing
from typing import TYPE_CHECKING  # isort:skip

if TYPE_CHECKING:
    from moat.lib.rpc import MsgSender
    from moat.lib.stream import BaseMsg

    from collections.abc import Awaitable
    from typing import Any


class MsgStream(HandlerStream):
    """
    This :moat.lib.rpc.stream:`HandlerStream` subclass
    interfaces with a `BaseCmdMsg` stream.

    """

    def __init__(self, handler: MsgSender, stream: BaseCmdMsg):
        self.__stream = stream
        super().__init__(handler)

    async def read_stream(self):
        "Background stream reader. Started from the HandlerStream context manager."
        str = self.__stream  # noqa: A001
        while True:
            msg = await str.recv()
            await self.msg_in(msg)

    async def write_stream(self):
        "Background stream writer. Started from the HandlerStream context manager."
        str = self.__stream  # noqa: A001
        while True:
            msg = await self.msg_out()
            await str.send(msg)


class BaseCmdMsg(BaseCmd):
    """
    This is a command handler that relays arbitrary MoaT-RPC messages
    and a `~moat.lib.stream.BaseMsg`-compatible stream.

    The difference to `~moat.lib.rpc.stream.cmdbbm.BaseCmdBBM` is that this
    class encapsulates arbitrary message/stream calls and requires a
    `~moat.lib.rpc.stream.cmdmsg.BaseCmdMsg` handler on the other side to talk to.

    In contrast, a `~moat.lib.rpc.stream.cmdbbm.BaseCmdBBM` exposes commands
    that directly read or write the underlying stream (of whatever type).

    This class cannot wrap a pre-existing stream. Its :meth:`stream` method
    **must** be overridden to create the stream.

    Parameters:
        prefix.recv(Path): Prefix for incoming messages
        prefix.send(Path): Prefix for outgoing messages
    """

    tg: TaskGroup = None
    __stream = None
    __rprefix = ()

    doc = dict(_d="Foo")

    async def stream(self) -> BaseMsg:
        """
        This method creates (and returns) the data stream.

        Must be overridden.

        Cleanup is typically handled via `moat.lib.micro.AC_use`.
        """
        raise NotImplementedError("Create the stream: ", self.__class__.__name__)

    async def setup(self):
        "Sets ``__rprefix``"
        await super().setup()

        rprefix = self.cfg.get("prefix", {}).get("send", ())
        if rprefix:
            rprefix = list(rprefix)
            rprefix.reverse()
            self.__rprefix = rprefix

    async def task(self):
        """
        Start the MsgStream.
        """
        root = self.root
        lprefix = self.cfg.get("prefix", {}).get("recv", ())
        if lprefix:
            root = root.sub_at(lprefix)
        try:
            self.s = await self.stream()
            async with MsgStream(root, self.s) as st:
                self.__stream = st
                if L:
                    self.set_ready()
                await idle()

        finally:
            self.s = None
            self.__stream = None

    async def reply_result(self, i, res):
        "send the result back"
        if i is None:
            return
        try:
            await self.s.send({"i": i, "d": res})
        except Exception as e:
            await self.reply_error(i, e)
        except BaseException as e:
            await self.reply_error(i, e)
            raise

    async def handle(self, msg, rcmd) -> Awaitable[Any]:
        """
        Forward a request to some remote side.
        """
        # Handle local commands (and documentation) locally
        if (
            (len(rcmd) == 1 or (len(rcmd) == 2 and rcmd[1] == "doc_"))
            and rcmd[0] != "dir_"
            and (hasattr(self, f"cmd_{rcmd[0]}") or hasattr(self, f"stream_{rcmd[0]}"))
        ):
            return await super().handle(msg, rcmd)

        if rcmd and rcmd[0] == "rdy_":
            if L and await self.wait_ready(wait=msg.get("wait", True)):
                raise NotReadyError(msg.cmd, rcmd)
        if self.__stream is None:
            raise EOFError

        # forward to remote
        rcmd.extend(self.__rprefix)
        res = await self.__stream.handle(msg, rcmd)

        # if it was a directory request, add local data
        if len(rcmd) == 1 and rcmd[0] == "dir_":
            # Merge local commands into the remote ones
            m2 = await self.cmd_dir_(v=msg.get("v", True))
            await msg.wait_replied(preload=True)

            msg._kw["c"] = tuple(set(msg.get("c", ())) | set(m2.pop("c", ())))  # noqa: SLF001
            msg._kw["s"] = tuple(set(msg.get("s", ())) | set(m2.pop("s", ())))  # noqa: SLF001
            merge(msg._kw, m2)  # noqa: SLF001
        return res

    doc_crd = dict(_d="read console", _0="int:len (64)")

    async def cmd_crd(self, n=64) -> bytes:
        """read some console data"""
        b = bytearray(n)
        if self.s is None:
            raise EOFError
        r = await self.s.crd(b)
        if r == n:
            return b
        elif r <= n >> 2:
            return bytes(b[:r])
        else:
            b = memoryview(b)
            return b[:r]

    doc_cwr = dict(_d="write console", _0="bytes:data")

    async def cmd_cwr(self, b):
        """write some console data"""
        if self.s is None:
            raise EOFError
        await self.s.cwr(b)

    doc_c = dict(
        _d="r/w console stream", _0="int:rdbuflen (64)", _i="bytes:to send", _o="bytes:received"
    )

    async def stream_c(self, msg):
        "read/write console stream"
        n = msg.get(0, 64)
        async with msg.stream() as st, TaskGroup() as tg:

            @tg.start_soon
            async def crd():
                while True:
                    await st.send(await self.cmd_crd(n))

            async for m in st:
                await self.cmd_cwr(m[0])
            tg.cancel()


class CmdMsg(BaseCmdMsg):
    """
    A baseCmdMsg with a ready-made link that it opens.
    """

    def __init__(self, link, cfg):
        super().__init__(cfg)
        self.link = link

    def stream(self) -> Awaitable[BaseMsg]:  # noqa:D102
        # pylint:disable=invalid-overridden-method
        return AC_use(self, self.link)


class SingleCmdMsg(BaseCmdMsg):
    """
    A BaseCmdMsg that disconnects on error, or when the connection ends,
    without propagating the exception.
    """

    # pylint:disable=abstract-method
    # `stream` needs to be implemented by a subclass

    async def run(self):  # noqa:D102
        # this would be far easier with "except*"
        # but µPy doesn't have that.
        try:
            try:
                await super().run()
            except BaseExceptionGroup as e:
                while True:
                    if len(e.exceptions) != 1:
                        a, b = e.split((EOFError, OSError, SilentRemoteError))
                        if a is not None:
                            log("Err %s: %r", self.path, repr(a))
                        if b is None:
                            return
                        raise b  # noqa:B904,RUF100
                    e = e.exceptions[0]
                    if not isinstance(e, BaseExceptionGroup):
                        raise e  # noqa:TRY201
        except EOFError:
            pass
        except (OSError, SilentRemoteError) as exc:
            log("Err %s: %r", self.path, repr(exc))
        except Exception as exc:  # pylint:disable=broad-exception-caught
            log("Err %s", self.path, err=exc)


class ExtCmdMsg(SingleCmdMsg):
    """SingleCmdMsg, on a stream that was established externally.

    The caller is responsible for calling :meth:`~moat.lib.rpc.BaseCmd.wait_stopped`
    and then closing the stream!
    """

    def __init__(self, stream: BaseMsg, cfg: dict[str, Any] | None = None):
        if cfg is None:
            cfg = {}
        super().__init__(cfg)
        self.__s = stream

    async def stream(self):  # noqa:D102
        return await AC_use(self, self.__s)



"""
Msghandler on top of anyio pipe
"""

from __future__ import annotations

import anyio
import logging
from contextlib import asynccontextmanager

from moat.util import ungroup

from .stream import HandlerStream

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from moat.lib.codec import Codec
    from moat.lib.rpc import MsgHandler, MsgSender

    from . import BaseMsgHandler

logger = logging.getLogger(__name__)

__all__ = ["AioStream", "rpc_on_aiostream"]


class AioStream(HandlerStream):  # noqa: D101
    __codec: Codec

    def __init__(
        self,
        cmd: MsgSender,
        stream,
        debug: str | None = None,
        codec: str | Codec | None = None,
        **kw,
    ):
        self.__s = stream
        self.__debug = debug

        if codec is None:
            from moat.lib.codec.moat_cbor import Codec as StdCBOR  # noqa: PLC0415

            codec = StdCBOR()

        elif isinstance(codec, str):
            from moat.lib.codec import get_codec  # noqa: PLC0415

            codec = get_codec(codec)

        self.__codec = codec
        super().__init__(cmd, **kw)

    async def read_stream(self):  # noqa: D102
        conn = self.__s
        codec = self.__codec
        rd_ = conn.read if hasattr(conn, "read") else conn.receive

        while True:
            buf = await rd_(4096)
            if self.__debug:
                logger.debug("R%s %r", self.__debug, buf)
            codec.feed(buf)
            for msg in codec:
                await self.msg_in(msg)

    async def write_stream(self):  # noqa: D102
        conn = self.__s
        codec = self.__codec
        wr = conn.write if hasattr(conn, "write") else conn.send
        while True:
            try:
                msg = await self.msg_out()
            except EOFError:
                return
            buf = codec.encode(msg)
            if self.__debug:
                logger.debug("W%s %r", self.__debug, bytes(buf))
            await wr(buf)


@asynccontextmanager
async def rpc_on_aiostream(
    cmd: BaseMsgHandler,
    stream: anyio.abc.ByteStream,
    *,
    codec: Codec | str | None = None,
    debug: bool = False,
    logger=None,
) -> MsgHandler:
    """
    Run a command handler on top of an anyio stream, using the given codec.

    Args:
        cmd: the handler for incoming messages. It may be `None`.

    This is an async context manager that yields the command handler.

    The default codec is :py:class:`moat.lib.codec.moat_cbor.Codec`.
    """

    y = False
    try:
        async with (
            ungroup,
            stream,
            AioStream(cmd, stream, codec=codec, debug=debug, logger=logger) as hs,
        ):
            y = True
            yield hs
    except (anyio.EndOfStream, anyio.BrokenResourceError, anyio.ClosedResourceError):
        if not y:
            raise
